/*--------------------------------*- C++ -*----------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Application
    fluentMeshToFoam

Description
    Converts a Fluent mesh to OpenFOAM format
    including multiple region and region boundary handling.

\*---------------------------------------------------------------------------*/

%{

#undef yyFlexLexer

/* ------------------------------------------------------------------------- *\
   ------ local definitions
\* ------------------------------------------------------------------------- */

#include "argList.H"
#include "Time.H"
#include "IStringStream.H"
#include "polyMesh.H"
#include "emptyPolyPatch.H"
#include "wallPolyPatch.H"
#include "cyclicAMIPolyPatch.H"
#include "symmetryPolyPatch.H"
#include "cellShape.H"
#include "faceSet.H"
#include "cellSet.H"
#include "meshTools.H"
#include "OFstream.H"
#include "wordIOList.H"

#include "readHexLabel.H"
#include "cellShapeRecognition.H"
#include "repatchPolyTopoChanger.H"

#include "edgeCollapser.H"
#include "PackedBoolList.H"

#include "plane.H"
#include "line.H"
#include "SortableList.H"
#include "SubField.H"

#include "mathematicalConstants.H"


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

using namespace Foam;

const scalar convertToMeters = 1.0;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //


label dimensionOfGrid = 0;

label nPoints = 0;
label nFaces = 0;
label nCells = 0;

pointField points(0);
SLList<label> pointGroupZoneID;
SLList<label> pointGroupStartIndex;
SLList<label> pointGroupEndIndex;


faceList faces(0);
labelList owner(0);
labelList neighbour(0);
SLList<label> faceGroupZoneID;
SLList<label> faceGroupStartIndex;
SLList<label> faceGroupEndIndex;

labelList fluentCellModelID(0);
SLList<label> cellGroupZoneID;
SLList<label> cellGroupStartIndex;
SLList<label> cellGroupEndIndex;
SLList<label> cellGroupType;

// number of zones adjusted at run-time if necessary
label maxZoneID = 100;
label zoneIDBuffer = 10;

wordList patchTypeIDs(maxZoneID);
wordList patchNameIDs(maxZoneID);

// Dummy yywrap to keep yylex happy at compile time.
// It is called by yylex but is not used as the mechanism to change file.
// See <<EOF>>
#if YY_FLEX_SUBMINOR_VERSION < 34
extern "C" int yywrap()
#else
int yyFlexLexer::yywrap()
#endif
{
    return 1;
}

%}

one_space                  [ \t\f]
space                      {one_space}*
some_space                 {one_space}+
cspace                     ","{space}
spaceNl                    ({space}|\n|\r)*

alpha                      [_[:alpha:]]
digit                      [[:digit:]]
decDigit                   [[:digit:]]
octalDigit                 [0-7]
hexDigit                   [[:xdigit:]]

lbrac                      "("
rbrac                      ")"
quote                      \"
dash                       "-"
dotColonDash               [.:-]

schemeSpecialInitial       [!$%&*/:<=>?~_^#.]
schemeSpecialSubsequent    [.+-]
schemeSymbol               (({some_space}|{alpha}|{quote}|{schemeSpecialInitial})({alpha}|{quote}|{digit}|{schemeSpecialInitial}|{schemeSpecialSubsequent})*)


identifier                 {alpha}({alpha}|{digit})*
integer                    {decDigit}+
label                      [1-9]{decDigit}*
hexLabel                   {hexDigit}+
zeroLabel                  {digit}*

word                       ({alpha}|{digit}|{dotColonDash})*

exponent_part              [eE][-+]?{digit}+
fractional_constant        [-+]?(({digit}*"."{digit}+)|({digit}+".")|({digit}))

double                     ((({fractional_constant}{exponent_part}?)|({digit}+{exponent_part}))|0)

x                          {double}
y                          {double}
z                          {double}
scalar                     {double}
labelListElement           {space}{zeroLabel}
hexLabelListElement        {space}{hexLabel}
scalarListElement          {space}{double}
schemeSymbolListElement    {space}{schemeSymbol}
labelList                  ({labelListElement}+{space})
hexLabelList               ({hexLabelListElement}+{space})
scalarList                 ({scalarListElement}+{space})
schemeSymbolList           ({schemeSymbolListElement}+{space})

starStar                   ("**")
text                       ({space}({word}*{space})*)

dateDDMMYYYY               ({digit}{digit}"/"{digit}{digit}"/"{digit}{digit}{digit}{digit})
dateDDMonYYYY              ((({digit}{digit}{space})|({digit}{space})){alpha}*{space}{digit}{digit}{digit}{digit})
time                       ({digit}{digit}":"{digit}{digit}":"{digit}{digit})

versionNumber              ({digit}|".")*

comment                    {spaceNl}"(0"{space}
header                     {spaceNl}"(1"{space}
dimension                  {spaceNl}"(2"{space}
point                      {spaceNl}"(10"{space}
fluentFace                 {spaceNl}"(13"{space}
cell                       {spaceNl}"(12"{space}
zoneVariant1               {spaceNl}"(39"{space}
zoneVariant2               {spaceNl}"(45"{space}

unknownPeriodicFace        {spaceNl}"(17"{space}
periodicFace               {spaceNl}"(18"{space}
cellTree                   {spaceNl}"(58"{space}
faceTree                   {spaceNl}"(59"{space}
faceParents                {spaceNl}"(61"{space}

endOfSection               {space}")"{space}



 /* ------------------------------------------------------------------------- *\
                      -----  Exclusive start states -----
 \* ------------------------------------------------------------------------- */

%option stack

%x readComment
%x embeddedCommentState
%x readHeader
%x readDimension
%x readPoint
%x readPointHeader
%x readNumberOfPoints
%x readPointGroupData
%x readPointData
%x readPoints2D
%x readPoints3D
%x fluentFace
%x readFaceHeader
%x readNumberOfFaces
%x readFaceGroupData
%x readFaceData
%x readFacesMixed
%x readFacesUniform
%x cell
%x readCellHeader
%x readNumberOfCells
%x readCellGroupData
%x readCellData
%x readCellsMixed
%x readCellsUniform
%x zone
%x readZoneHeader
%x readZoneGroupData
%x readZoneData
%x readZoneBlock

%x periodicFace
%x cellTree
%x faceTree
%x faceParents

%x unknownBlock
%x embeddedUnknownBlock
%%

%{
    // Point data
    label pointGroupNumberOfComponents = 3;
    label pointI = 0; // index used for reading points

    // Face data
    label faceGroupElementType = -1;
    label faceI = 0;

    // Cell data
    label cellGroupElementType = -1;
    label celli = 0;
%}


 /* ------------------------------------------------------------------------- *\
                            ------ Start Lexing ------
 \* ------------------------------------------------------------------------- */

 /*                      ------ Reading control header ------                 */

{comment} {
        yy_push_state(readComment);
    }


<readComment>{quote}{text}{quote} {
    }


<readComment>{spaceNl}{endOfSection} {
        yy_pop_state();
    }

{header} {
        BEGIN(readHeader);
    }

<readHeader>{quote}{text}{quote} {
        Info<< "Reading header: " << YYText() << endl;
    }


{dimension} {
        BEGIN(readDimension);
    }

<readDimension>{space}{label}{space} {
        IStringStream dimOfGridStream(YYText());

        dimensionOfGrid = readLabel(dimOfGridStream);

        Info<< "Dimension of grid: " << dimensionOfGrid << endl;
    }


{point} {
        BEGIN(readPointHeader);
    }

<readPointHeader>{spaceNl}{lbrac}{space}"0"{space}"1"{space} {
        BEGIN(readNumberOfPoints);
    }

<readNumberOfPoints>{space}{hexLabel}{space}{labelList} {

        IStringStream numberOfPointsStream(YYText());

        nPoints = readHexLabel(numberOfPointsStream);

        Info<< "Number of points: " << nPoints << endl;

        // set the size of the points list
        points.setSize(nPoints);

        // meaningless type skipped
        readLabel(numberOfPointsStream);

        // this dimension of grid may be checked against global dimension
        if (numberOfPointsStream)
        {
            // check dimension of grid
            readLabel(numberOfPointsStream);
        }
        else
        {
            Info<< endl;
        }
    }

<readPointHeader>{spaceNl}{lbrac} {
        BEGIN(readPointGroupData);
    }

<readPointGroupData>{space}{hexLabel}{space}{hexLabel}{space}{hexLabel}{labelList} {
        IStringStream pointGroupDataStream(YYText());

        // read point zone-ID, start and end-label
        // the indices will be used for checking later.
        pointGroupZoneID.append(readHexLabel(pointGroupDataStream));

        pointGroupStartIndex.append(readHexLabel(pointGroupDataStream));

        pointGroupEndIndex.append(readHexLabel(pointGroupDataStream));

        // point group type skipped
        readHexLabel(pointGroupDataStream);

        // In FOAM, indices start from zero - adjust
        pointI = pointGroupStartIndex.last() - 1;

        // reset number of components to default
        pointGroupNumberOfComponents = 3;

        // read number of components in the vector
        if (pointGroupDataStream)
        {
            pointGroupNumberOfComponents = readLabel(pointGroupDataStream);
        }
    }

<readNumberOfPoints,readPointGroupData>{endOfSection} {
        BEGIN(readPointData);
    }

<readPointData>{spaceNl}{lbrac} {

        Info<< "Reading points" << endl;

        if (pointGroupNumberOfComponents == 2)
        {
            yy_push_state(readPoints2D);
        }
        else
        {
            yy_push_state(readPoints3D);
        }
    }

<readPoints2D>{spaceNl}{scalarList} {

        IStringStream vertexXyzStream(YYText());

        // Note: coordinates must be read one at the time.
        scalar x = readScalar(vertexXyzStream);
        scalar y = readScalar(vertexXyzStream);

        points[pointI] = point(x, y, 0);
        pointI++;
    }

<readPoints3D>{spaceNl}{scalarList} {

        IStringStream vertexXyzStream(YYText());

        // Note: coordinates must be read one at the time.
        scalar x = readScalar(vertexXyzStream);
        scalar y = readScalar(vertexXyzStream);
        scalar z = readScalar(vertexXyzStream);

        points[pointI] = convertToMeters*point(x, y, z);
        pointI++;
    }

<readPoints2D,readPoints3D>{spaceNl}{endOfSection} {

        // check read of points
        if (pointI != pointGroupEndIndex.last())
        {
            Info<< "problem with reading points: "
                << "start index: " << pointGroupStartIndex.last()
                << " end index: " << pointGroupEndIndex.last()
                << " last points read: " << pointI << endl;
        }

        yy_pop_state();
    }


{fluentFace} {
        BEGIN(readFaceHeader);
    }

<readFaceHeader>{spaceNl}{lbrac}{space}"0"{space}"1"{space} {
        BEGIN(readNumberOfFaces);
    }

<readNumberOfFaces>{space}{hexLabel}{space}{labelListElement}+ {

        IStringStream numberOfFacesStream(YYText());

        nFaces = readHexLabel(numberOfFacesStream);

        Info<< "number of faces: " << nFaces << endl;

        faces.setSize(nFaces);
        owner.setSize(nFaces);
        neighbour.setSize(nFaces);

        // Meaningless type and element type not read
    }

<readFaceHeader>{spaceNl}{lbrac} {
        BEGIN(readFaceGroupData);
    }

<readFaceGroupData>{space}{hexLabel}{space}{hexLabel}{space}{hexLabel}{hexLabelListElement}+ {

        IStringStream faceGroupDataStream(YYText());

        // read fluentFace zone-ID, start and end-label
        faceGroupZoneID.append(readHexLabel(faceGroupDataStream));

        // the indices will be used for checking later.
        faceGroupStartIndex.append(readHexLabel(faceGroupDataStream));

        faceGroupEndIndex.append(readHexLabel(faceGroupDataStream));

        // face group type
        readHexLabel(faceGroupDataStream);

        faceGroupElementType = readHexLabel(faceGroupDataStream);

        // In FOAM, indices start from zero - adjust
        faceI = faceGroupStartIndex.last() - 1;
    }

<readNumberOfFaces,readFaceGroupData>{spaceNl}{endOfSection} {
        BEGIN(readFaceData);
    }

<readFaceData>{spaceNl}{lbrac} {

        if (faceGroupElementType == 0)
        {
            Info<< "Reading mixed faces" << endl;
            yy_push_state(readFacesMixed);
        }
        else
        {
            Info<< "Reading uniform faces" << endl;
            yy_push_state(readFacesUniform);
        }
    }

<readFacesMixed>{spaceNl}{hexLabelList} {

        IStringStream mixedFaceStream(YYText());

        face& curFaceLabels = faces[faceI];

        // set size of label list
        curFaceLabels.setSize(readLabel(mixedFaceStream));

        forAll (curFaceLabels, i)
        {
            curFaceLabels[i] = readHexLabel(mixedFaceStream) - 1;
        }

        // read neighbour and owner. Neighbour comes first
        neighbour[faceI] = readHexLabel(mixedFaceStream) - 1;
        owner[faceI] = readHexLabel(mixedFaceStream) - 1;
        faceI++;
    }

<readFacesUniform>{spaceNl}{hexLabelList} {

        IStringStream mixedFaceStream(YYText());

        face& curFaceLabels = faces[faceI];

        // set size of label list. This is OK because in Fluent the type
        // for edge is 2, for triangle is 3 and for quad is 4
        curFaceLabels.setSize(faceGroupElementType);

        forAll (curFaceLabels, i)
        {
            curFaceLabels[i] = readHexLabel(mixedFaceStream) - 1;
        }

        // read neighbour and owner. Neighbour comes first
        neighbour[faceI] = readHexLabel(mixedFaceStream) - 1;
        owner[faceI] = readHexLabel(mixedFaceStream) - 1;
        faceI++;
    }

<readFacesMixed,readFacesUniform>{spaceNl}{endOfSection} {

        // check read of fluentFaces
        if (faceI != faceGroupEndIndex.last())
        {
            Info<< "problem with reading fluentFaces: "
                << "start index: " << faceGroupStartIndex.last()
                << " end index: " << faceGroupEndIndex.last()
                << " last fluentFaces read: " << faceI << endl;
        }

        yy_pop_state();
    }


{cell} {
        BEGIN(readCellHeader);
    }

<readCellHeader>{spaceNl}{lbrac}{space}"0"{space}"1"{space} {
        BEGIN(readNumberOfCells);
    }

<readNumberOfCells>{space}{hexLabel}{space}{labelListElement}+ {

        IStringStream numberOfCellsStream(YYText());

        nCells = readHexLabel(numberOfCellsStream);

        Info<< "Number of cells: " << nCells << endl;

        fluentCellModelID.setSize(nCells);

        // Meaningless type and element type not read
    }

<readCellHeader>{spaceNl}{lbrac} {
        BEGIN(readCellGroupData);
    }

<readCellGroupData>{space}{hexLabel}{space}{hexLabel}{space}{hexLabel}{space}{hexLabel} {
        // Warning. This entry must be above the next one because of the lexing
        // rules. It is introduced to deal with the problem of reading
        // non-standard cell definition from Tgrid, which misses the type label.

        Info<< "Tgrid syntax problem: " << YYText() << endl;
        IStringStream cellGroupDataStream(YYText());

        // read cell zone-ID, start and end-label
        cellGroupZoneID.append(readHexLabel(cellGroupDataStream));

        // the indices will be used for checking later.
        cellGroupStartIndex.append(readHexLabel(cellGroupDataStream));

        cellGroupEndIndex.append(readHexLabel(cellGroupDataStream));

        cellGroupType.append(readHexLabel(cellGroupDataStream));

        Info<< "cellGroupZoneID:" << cellGroupZoneID.last()
            << endl;
        Info<< "cellGroupStartIndex:" << cellGroupStartIndex.last()
            << endl;
        Info<< "cellGroupEndIndex:" << cellGroupEndIndex.last()
            << endl;
        Info<< "cellGroupType:" << cellGroupType.last()
            << endl;


        // Note. Potentially skip cell set if type is zero.
        // This entry does not exist in Tgrid files.
        if (dimensionOfGrid == 2)
        {
            // Tgrid creating triangles
            cellGroupElementType = 1;
        }
        else
        {
            cellGroupElementType = 2;
        }

        // In FOAM, indices start from zero - adjust
        celli = cellGroupStartIndex.last() - 1;

        if (cellGroupElementType != 0)
        {
            label lastIndex = cellGroupEndIndex.last();

            for (; celli < lastIndex; celli++)
            {
                fluentCellModelID[celli] = cellGroupElementType;
            }
        }
    }

<readCellGroupData>{space}{hexLabel}{space}{hexLabel}{space}{hexLabel}{space}{hexLabel}{space}{hexLabel} {
        // Warning. See above

        Info<< "Other readCellGroupData: " << YYText() << endl;


        IStringStream cellGroupDataStream(YYText());

        // read cell zone-ID, start and end-label
        cellGroupZoneID.append(readHexLabel(cellGroupDataStream));

        // the indices will be used for checking later.
        cellGroupStartIndex.append(readHexLabel(cellGroupDataStream));

        cellGroupEndIndex.append(readHexLabel(cellGroupDataStream));

        cellGroupType.append(readHexLabel(cellGroupDataStream));

        // Note. Potentially skip cell set if type is zero.

        cellGroupElementType = readHexLabel(cellGroupDataStream);

        // In FOAM, indices start from zero - adjust
        celli = cellGroupStartIndex.last() - 1;

        if (cellGroupElementType != 0)
        {
            Info<< "Reading uniform cells" << endl;
            label lastIndex = cellGroupEndIndex.last();

            for (; celli < lastIndex; celli++)
            {
                fluentCellModelID[celli] = cellGroupElementType;
            }
        }
    }

<readNumberOfCells,readCellGroupData>{endOfSection} {
        BEGIN(readCellData);
    }

<readCellData>{spaceNl}{lbrac} {
        Info<< "Reading mixed cells" << endl;
        yy_push_state(readCellsMixed);
    }

<readCellsMixed>{spaceNl}{labelList} {

        IStringStream fluentCellModelIDStream(YYText());

        label celliD;
        while (fluentCellModelIDStream.read(celliD))
        {
            fluentCellModelID[celli] = celliD;
            celli++;
        }
    }

<readCellsMixed>{spaceNl}{endOfSection} {

        // check read of cells
        if (celli != cellGroupEndIndex.last())
        {
            Info<< "problem with reading cells: "
                << "start index: " << cellGroupStartIndex.last()
                << " end index: " << cellGroupEndIndex.last()
                << " last cells read: " << celli << endl;
        }

        yy_pop_state();
    }



{zoneVariant1} {
        BEGIN(readZoneHeader);
    }

{zoneVariant2} {
        BEGIN(readZoneHeader);
    }

<readZoneHeader>{spaceNl}{lbrac} {
        BEGIN(readZoneGroupData);
    }

<readZoneGroupData>{space}{label}{space}{word}{space}{word} {

        IStringStream zoneDataStream(YYText());

        // cell zone-ID not in hexadecimal!!! Inconsistency
        label zoneID(readLabel(zoneDataStream));

        if (zoneID > maxZoneID - 1)
        {
            // resize the container
            maxZoneID = zoneID + zoneIDBuffer;

            patchTypeIDs.setSize(maxZoneID);
            patchNameIDs.setSize(maxZoneID);
        }

        zoneDataStream >> patchTypeIDs[zoneID];
        zoneDataStream >> patchNameIDs[zoneID];

        Info<< "Read zone1:" << zoneID
            << " name:" << patchNameIDs[zoneID]
            << " patchTypeID:" << patchTypeIDs[zoneID]
            << endl;
    }

<readZoneGroupData>{space}{label}{space}{word}{space}{word}{space}{label} {
        // Fluent manual inconsistency, version 6.1.22
        IStringStream zoneDataStream(YYText());

        // cell zone-ID not in hexadecimal!!! Inconsistency
        label zoneID(readLabel(zoneDataStream));

        if (zoneID > maxZoneID - 1)
        {
            // resize the container
            maxZoneID = zoneID + zoneIDBuffer;

            patchTypeIDs.setSize(maxZoneID);
            patchNameIDs.setSize(maxZoneID);
        }

        zoneDataStream >> patchTypeIDs[zoneID];
        zoneDataStream >> patchNameIDs[zoneID];

        Info<< "Read zone2:" << zoneID
            << " name:" << patchNameIDs[zoneID]
            << " patchTypeID:" << patchTypeIDs[zoneID]
            << endl;
    }

<readZoneGroupData>{endOfSection} {
        BEGIN(readZoneData);
    }

<readZoneData>{spaceNl}{lbrac} {
        Info<< "Reading zone data" << endl;
        yy_push_state(readZoneBlock);
    }

<readZoneBlock>{spaceNl}{schemeSymbolList} {
    }

<readZoneBlock>{lbrac} {
        yy_push_state(unknownBlock);
    }

<readZoneBlock>{endOfSection} {
        yy_pop_state();
    }



 /*             ------ Reading end of section and others ------               */

<readHeader,readDimension,readPointData,readFaceData,readCellData,readZoneData>{spaceNl}{endOfSection} {
        BEGIN(INITIAL);
    }

 /*    ------ Reading unknown type or non-standard comment ------             */


{lbrac}{label} {
        yy_push_state(unknownBlock);
    }

<readComment,unknownBlock,embeddedUnknownBlock>{spaceNl}{schemeSymbol} {
    }

<readComment,unknownBlock,embeddedUnknownBlock>{spaceNl}{lbrac} {
        yy_push_state(embeddedUnknownBlock);

    }

<readComment,unknownBlock,embeddedUnknownBlock>{spaceNl}{endOfSection} {
        yy_pop_state();
    }

<unknownBlock,embeddedUnknownBlock>{spaceNl}{labelList} {
    }

<unknownBlock,embeddedUnknownBlock>{spaceNl}{hexLabelList} {
    }

<unknownBlock,embeddedUnknownBlock>{spaceNl}{scalarList} {
    }

<unknownBlock,embeddedUnknownBlock>{spaceNl}{schemeSymbolList} {
    }

<unknownBlock,embeddedUnknownBlock>{spaceNl}{text} {
    }


 /* ------ Ignore remaining space and \n s.  Any other characters are errors. */

<readPoints2D,readPoints3D>.|\n {
    Info<< "ERROR! Do not understand characters: " << YYText() << endl;
    }
.|\n {}


 /*  ------ On EOF return to previous file, if none exists terminate. ------  */

<<EOF>> {
            yyterminate();
    }
%%


#include "fileName.H"
#include <fstream>

using namespace Foam;

typedef line<point,point> linie;

// get the plane of the grid
// right now it assumes parallel to XY

plane getPlane(pointField& points) 
{
    scalar x=0,y=0,z=0;
    label  n=0;

    forAll(points,pointI) 
    {
        point &pt=points[pointI];
        x += pt.x();    
        y += pt.y();    
        z += pt.z();
        n ++;
    }

    plane result(point(x/n,y/n,z/n),vector(0,0,1));

    return result;
}

// calculate the symmetry axis from the "symmetry patch"
// by finding the two points that are farthest apart

linie getAxis(const polyPatch &axis,plane &pl) 
{
    scalar dist=0;
    int cnt=0;
    point pt1(0,0,0),pt2(0,0,0);
    const pointField &pts=axis.localPoints();

    forAll(pts,pointI) 
    {
        switch(cnt) 
        {
            case 0:
                pt1=pl.nearestPoint(pts[pointI]);
                break;

            case 1:
                pt2=pl.nearestPoint(pts[pointI]);
                dist=mag(pt1-pt2);
                break;

            default:
                point pt=pl.nearestPoint(pts[pointI]);

            if( mag(pt-pt1)>dist || mag(pt-pt2)>dist) 
            {
	        if( mag(pt-pt1) > mag(pt-pt2)) 
                {
	            pt2=pt;
	        } 
                else  
                {
	            pt1=pt;
	        }
	        dist=mag(pt1-pt2);
            }
        }
        cnt++;
    }

    return linie(pt1,pt2);
}

scalar getDistance(const polyPatch &axisPatch,linie &axisLine) 
{
    const pointField &pts=axisPatch.localPoints();
    vector axisDir = axisLine.vec()/axisLine.mag();      

    scalar distance=0;

    forAll(pts,pointI) 
    {
        const point &pt=pts[pointI];

        point axisPoint = axisLine.start() + axisDir * (axisDir & (pt-axisLine.start()));
        scalar radius=mag( axisPoint - pt );

        if(radius>distance) 
        {
            distance=radius;
        }
    }
  
    return distance;
}

const scalar defaultAngle=5;

void changeCoordinates
(
    polyMesh &mesh,
    plane cutPlane,
    linie &axisLine,
    scalar offset,
    const scalar wedgeAngle,
    bool revolve
) 
{
    if(revolve)
    {
        Info << "Revolving nodes" << endl;
    }
    else
    {
        Info << "Projecting nodes" << endl;
    }

    const scalar angle=wedgeAngle/2;

    repatchPolyTopoChanger topo(mesh);

    pointField oldPoints=mesh.points();//DPS the old points will be rotated too

    pointField newPoints(oldPoints.size());

    const scalar factor=std::sin(angle/180.*constant::mathematical::pi);
    const scalar factorRadius=std::cos(angle/180.*constant::mathematical::pi);
    vector axisDir = axisLine.vec()/axisLine.mag();      

    scalar minRadius=1e10,maxRadius=-1e10;

    scalar dist1=mag(axisLine.start()-cutPlane.nearestPoint(axisLine.start()));
    scalar dist2=mag(axisLine.end()-cutPlane.nearestPoint(axisLine.end()));

    if(dist1>SMALL || dist2>SMALL)  
    {
        Warning << " End points of axis " << axisLine << " are " << dist1 
                << " and " << dist2 << " away from plane "  << cutPlane << endl;
    }

    forAll(oldPoints,pointI) 
    {
        //    point oldPoint=oldPoints[pointI];
        const point &oldPoint=oldPoints[pointI];
        point basePoint=cutPlane.nearestPoint(oldPoint);

        scalar radius=GREAT;

        point axisPoint = axisLine.start() + axisDir * (axisDir & (basePoint-axisLine.start()));
        radius=mag( axisPoint - basePoint )+offset;
        point radiusBasedPoint = factorRadius * (basePoint - axisPoint);

        if(radius>maxRadius) 
        {
            maxRadius=radius;
        }

        if(radius<minRadius) 
        {
            minRadius=radius;
        }

        vector dir(oldPoint-basePoint);

        if(revolve) 
        {
            newPoints[pointI] 
                = axisPoint + radiusBasedPoint + factor*radius*dir/mag(dir);
        }
        else newPoints[pointI] = basePoint + factor * radius * dir/mag(dir);
        //    oldPoints[pointI] = basePoint - factor *radius * dir/mag(dir);  //DPS rotating original plane
    }

    Info << "Radius to axis: min = " << minRadius << " max = " << maxRadius << endl;

    //  mesh.movePoints(oldPoints);
    mesh.movePoints(newPoints);
}

// Split the wedge-patch into two patches
 
void splitWedge(polyMesh &mesh,word wname,plane pl) 
{
    repatchPolyTopoChanger topo(mesh);

    const polyBoundaryMesh& patches = mesh.boundaryMesh();
    const polyPatch &wedge=patches[patches.findPatchID(wname)];
  
    const vectorField::subField 	& fcs=wedge.faceCentres ();

    faceSet facesPos(mesh,"set1_"+wname,fcs.size()/2,IOobject::NO_WRITE);
    faceSet facesNeg(mesh,"set2_"+wname,fcs.size()/2,IOobject::NO_WRITE);

    // assign faces to the two patches
    forAll(fcs,faceI) 
    {
        vector center=fcs[faceI];
        scalar dir=pl.normal() & (center-pl.nearestPoint(center));
 
        if( dir >0 ) 
        {
            // This offset is dirty, there must be a better way
            facesPos.insert(faceI-wedge.whichFace(0));
        } 
        else 
        {
            facesNeg.insert(faceI-wedge.whichFace(0));
        }
    }

    Info << " Copying patches " << endl;

    // the rest is stolen from createPatch (others might say inspired by)

    List<polyPatch*> newPatches(patches.size() + 2);
  
    forAll(patches, patchI)
    {
        const polyPatch& pp = patches[patchI];
      
        newPatches[patchI] =
	    pp.clone
	    (
	        patches,
	        patchI,
	        pp.size(),
	        pp.start()
	    ).ptr();
    }
  
    label patchPos = newPatches.size() - 2;
    label patchNeg = newPatches.size() - 1;
  
    Info << " Creating Patches" << endl;

    newPatches[patchPos] =
        polyPatch::New
        (
            "empty",
            "wedge_front",
            0,
            mesh.nFaces(),
            patchPos,
            patches
        ).ptr();

    newPatches[patchNeg] =
        polyPatch::New
        (
            "empty",
            "wedge_back",
            0,
            mesh.nFaces(),
            patchNeg,
            patches
        ).ptr();
  

    // Actually add new list of patches
    topo.changePatches(newPatches);
  
    Info << " Creating Pos-patch " << endl;

    labelList faceLabelsPos(facesPos.toc());

    SortableList<label> patchFacesPos(faceLabelsPos);

    forAll(patchFacesPos, i)
    {
        label faceI = patchFacesPos[i];

        if (mesh.isInternalFace(faceI))
	{
	    FatalErrorIn("SplitWedge")
	        << "Face " << faceI << " specified in set " << facesPos.name()
	        << " is not an external face of the mesh." << endl
	        << "This application can only repatch existing boundary"
	        << " faces."
	        << exit(FatalError);
	}
        topo.changePatchID(patchFacesPos[i], patchPos);
    }

    Info << " Creating Neg-patch " << endl;

    labelList faceLabelsNeg(facesNeg.toc());

    SortableList<label> patchFacesNeg(faceLabelsNeg);

    forAll(patchFacesNeg, i)
    {
        label faceI = patchFacesNeg[i];

        if (mesh.isInternalFace(faceI))
	{
	    FatalErrorIn("SplitWedge")
	        << "Face " << faceI << " specified in set " << facesNeg.name()
	        << " is not an external face of the mesh." << endl
	        << "This application can only repatch existing boundary"
	        << " faces."
	        << exit(FatalError);
	}
        topo.changePatchID(patchFacesNeg[i], patchNeg);
    }

    Info << " Changing patches\n" << endl;

    topo.repatch();
}

void changeTypes(polyMesh &mesh,word wedge,word axis,bool hasOffset) 
{
    repatchPolyTopoChanger topo(mesh);
    const polyBoundaryMesh& patches = mesh.boundaryMesh();

    List<polyPatch*> newPatches(patches.size());
  
    forAll(patches, patchI)
    {
        const polyPatch& pp = patches[patchI];

        if(pp.name()==axis && !hasOffset) 
        {
            Info << " Changing " << axis << " to empty" << endl;

            newPatches[patchI] =
                polyPatch::New
                (
                    "empty",
                    pp.name(),
                    pp.size(),
                    pp.start(),
                    patchI,
                    patches
                ).ptr();
        } 
        else if(pp.name()=="wedge_front" || pp.name()=="wedge_back")
        {
            Info << " Changing " << pp.name() << " to wedge" << endl;

            newPatches[patchI] =
                polyPatch::New
                (
                    "wedge",
                    pp.name(),
                    pp.size(),
                    pp.start(),
                    patchI,
                    patches
                ).ptr();
        } 
        else 
        {
            newPatches[patchI] =
                pp.clone
                (
                    patches,
                    patchI,
                    pp.size(),
                    pp.start()
                ).ptr();
        }
    }   

    Info << endl;

    topo.changePatches(newPatches);
}


// Find label of face.
label findFace(const primitiveMesh& mesh, const face& f)
{
    // Get faces using zeroth vertex of face.
    const labelList& pFaces = mesh.pointFaces()[f[0] ];

    forAll(pFaces, i)
    {
        label faceI = pFaces[i];

        if (f == mesh.faces()[faceI])
        {
            return faceI;
        }
    }

    // Didn't find face. Do what?
    FatalErrorIn("findFace(const primitiveMesh&, const face&)")
        << "Problem : cannot find a single face in the mesh which uses"
        << " vertices " << f << exit(FatalError);

    return -1;
}


labelList getSortedEdges
(
    const edgeList& edges,
    const labelList& f,
    const labelList& edgeLabels
)
{
    labelList faceEdges(edgeLabels.size(), -1);

    // Find starting pos in f for every edgeLabels
    forAll(edgeLabels, i)
    {
        label edgeI = edgeLabels[i];

        const edge& e = edges[edgeI];

        label fp = findIndex(f, e[0]);
        label fp1 = f.fcIndex(fp);

        if (f[fp1] == e[1])
        {
            // EdgeI between fp -> fp1
            faceEdges[fp] = edgeI;
        }
        else
        {
            // EdgeI between fp-1 -> fp
            faceEdges[f.rcIndex(fp)] = edgeI;
        }
    }

    return faceEdges;
}


// Merges edges which are in straight line. I.e. edge split by point.
label mergeEdges
(
    const polyMesh& mesh,
    const scalar maxCos,
    edgeCollapser& collapser
)
{
    const pointField& points = mesh.points();
    const edgeList& edges = mesh.edges();
    const labelListList& pointEdges = mesh.pointEdges();
    const labelList& region = collapser.pointRegion();
    const labelList& master = collapser.pointRegionMaster();

    label nCollapsed = 0;

    forAll(pointEdges, pointI)
    {
        const labelList& pEdges = pointEdges[pointI];

        if (pEdges.size() == 2)
        {
            const edge& leftE = edges[pEdges[0]];
            const edge& rightE = edges[pEdges[1]];

            // Get the two vertices on both sides of the point
            label leftV = leftE.otherVertex(pointI);
            label rightV = rightE.otherVertex(pointI);

            // Collapse only if none of the points part of merge network
            // or all of networks with different masters.
            label midMaster = -1;
            if (region[pointI] != -1)
            {
                midMaster = master[region[pointI]];
            }

            label leftMaster = -2;
            if (region[leftV] != -1)
            {
                leftMaster = master[region[leftV]];
            }

            label rightMaster = -3;
            if (region[rightV] != -1)
            {
                rightMaster = master[region[rightV]];
            }

            if
            (
                midMaster != leftMaster
             && midMaster != rightMaster
             && leftMaster != rightMaster
            )
            {
                // Check if the two edge are in line
                vector leftVec = points[pointI] - points[leftV];
                leftVec /= mag(leftVec) + VSMALL;

                vector rightVec = points[rightV] - points[pointI];
                rightVec /= mag(rightVec) + VSMALL;

                if ((leftVec & rightVec) > maxCos)
                {
                    // Collapse one (left) side of the edge. Make left vertex
                    // the master.
                    //if (collapser.unaffectedEdge(pEdges[0]))
                    {
                        collapser.collapseEdge(pEdges[0], leftV);
                        nCollapsed++;
                    }
                }
            }
        }
    }

    return nCollapsed;
}


// Return master point edge needs to be collapsed to (or -1)
label edgeMaster(const PackedBoolList& boundaryPoint, const edge& e)
{
    label masterPoint = -1;

    // Collapse edge to boundary point.
    if (boundaryPoint.get(e[0]))
    {
        if (boundaryPoint.get(e[1]))
        {
            // Both points on boundary. Choose one to collapse to.
            // Note: should look at feature edges/points!
            masterPoint = e[0];
        }
        else
        {
            masterPoint = e[0];
        }
    }
    else
    {
        if (boundaryPoint.get(e[1]))
        {
            masterPoint = e[1];
        }
        else
        {
            // None on boundary. Choose arbitrary.
            // Note: should look at geometry?
            masterPoint = e[0];
        }
    }
    return masterPoint;
}


label collapseSmallEdges
(
    const polyMesh& mesh,
    const PackedBoolList& boundaryPoint,
    const scalar minLen,
    edgeCollapser& collapser
)
{
    const pointField& points = mesh.points();
    const edgeList& edges = mesh.edges();

    // Collapse all edges that are too small. Choose intelligently which
    // point to collapse edge to.

    label nCollapsed = 0;

    forAll(edges, edgeI)
    {
        const edge& e = edges[edgeI];

        if (e.mag(points) < minLen)
        {
            label master = edgeMaster(boundaryPoint, e);

            if (master != -1) // && collapser.unaffectedEdge(edgeI))
            {
                collapser.collapseEdge(edgeI, master);
                nCollapsed++;
            }
        }
    }
    return nCollapsed;
}


// Faces which have edges just larger than collapse length but faces which
// are very small. This one tries to collapse them if it can be done with
// edge collapse. For faces where a face gets replace by two edges use
// collapseFaces
label collapseHighAspectFaces
(
    const polyMesh& mesh,
    const PackedBoolList& boundaryPoint,
    const scalar areaFac,
    const scalar edgeRatio,
    edgeCollapser& collapser
)
{
    const pointField& points = mesh.points();
    const edgeList& edges = mesh.edges();
    const faceList& faces = mesh.faces();
    const labelListList& faceEdges = mesh.faceEdges();

    scalarField magArea(mag(mesh.faceAreas()));

    label maxIndex = findMax(magArea);

    scalar minArea = areaFac * magArea[maxIndex];

    Info<< "Max face area:" << magArea[maxIndex] << endl
        << "Collapse area factor:" << areaFac << endl
        << "Collapse area:" << minArea << endl;

    label nCollapsed = 0;

    forAll(faces, faceI)
    {
        if (magArea[faceI] < minArea)
        {
            const face& f = faces[faceI];

            // Get the edges in face point order
            labelList fEdges(getSortedEdges(edges, f, faceEdges[faceI]));

            SortableList<scalar> lengths(fEdges.size());
            forAll(fEdges, i)
            {
                lengths[i] = edges[fEdges[i]].mag(points);
            }
            lengths.sort();


            label edgeI = -1;

            if (f.size() == 4)
            {
                // Compare second largest to smallest
                if (lengths[2] > edgeRatio*lengths[0])
                {
                    // Collapse smallest only. Triangle should be cleared
                    // next time around.
                    edgeI = fEdges[lengths.indices()[0]];
                }
            }
            else if (f.size() == 3)
            {
                // Compare second largest to smallest
                if (lengths[1] > edgeRatio*lengths[0])
                {
                    edgeI = fEdges[lengths.indices()[0]];
                }
            }


            if (edgeI != -1)
            {
                label master = edgeMaster(boundaryPoint, edges[edgeI]);

                if (master != -1)// && collapser.unaffectedEdge(edgeI))
                {
                    collapser.collapseEdge(edgeI, master);
                    nCollapsed++;
                }
            }
        }
    }

    return nCollapsed;
}


void set(const labelList& elems, const bool val, boolList& status)
{
    forAll(elems, i)
    {
        status[elems[i]] = val;
    }
}


// Tries to simplify polygons to face of minSize (4=quad, 3=triangle)
label simplifyFaces
(
    const polyMesh& mesh,
    const PackedBoolList& boundaryPoint,
    const label minSize,
    const scalar lenGap,
    edgeCollapser& collapser
)
{
    const pointField& points = mesh.points();
    const edgeList& edges = mesh.edges();
    const faceList& faces = mesh.faces();
    const cellList& cells = mesh.cells();
    const labelListList& faceEdges = mesh.faceEdges();
    const labelList& faceOwner = mesh.faceOwner();
    const labelList& faceNeighbour = mesh.faceNeighbour();
    const labelListList& pointCells = mesh.pointCells();
    const labelListList& cellEdges = mesh.cellEdges();

    label nCollapsed = 0;

    boolList protectedEdge(mesh.nEdges(), false);

    forAll(faces, faceI)
    {
        const face& f = faces[faceI];

        if
        (
            f.size() > minSize
         && cells[faceOwner[faceI]].size() >= 6
         && (
                mesh.isInternalFace(faceI)
             && cells[faceNeighbour[faceI]].size() >= 6
            )
        )
        {
            // Get the edges in face point order
            labelList fEdges(getSortedEdges(edges, f, faceEdges[faceI]));

            SortableList<scalar> lengths(fEdges.size());
            forAll(fEdges, i)
            {
                lengths[i] = edges[fEdges[i]].mag(points);
            }
            lengths.sort();


            // Now find a gap in length between consecutive elements greater
            // than lenGap.

            label gapPos = -1;

            for (label i = f.size()-1-minSize; i >= 0; --i)
            {
                if (lengths[i+1] > lenGap*lengths[i])
                {
                    gapPos = i;

                    break;
                }
            }

            if (gapPos != -1)
            {
                //for (label i = gapPos; i >= 0; --i)
                label i = 0;  // Hack: collapse smallest edge only.
                {
                    label edgeI = fEdges[lengths.indices()[i]];

                    if (!protectedEdge[edgeI])
                    {
                        const edge& e = edges[edgeI];

                        label master = edgeMaster(boundaryPoint, e);

                        if (master != -1)
                        {
                            collapser.collapseEdge(edgeI, master);

                            // Protect all other edges on all cells using edge
                            // points.

                            const labelList& pCells0 = pointCells[e[0]];

                            forAll(pCells0, i)
                            {
                                set(cellEdges[pCells0[i]], true, protectedEdge);
                            }
                            const labelList& pCells1 = pointCells[e[1]];

                            forAll(pCells1, i)
                            {
                                set(cellEdges[pCells1[i]], true, protectedEdge);
                            }

                            nCollapsed++;
                        }
                    }
                }
            }
        }
    }

    return nCollapsed;
}

void makeAxialMesh(polyMesh& mesh, fileName& axisName)
{
    Info<< "Create axial mesh" << nl << endl;

    word axisName_(axisName);
    word wedgeName;

    bool revolve = false;

    scalar offset=0;
    scalar wedgeAngle=defaultAngle;

    vector rotation(1,0,0),origin(0,0,0);

    wedgeName = "frontAndBackPlanes";
    wedgeAngle = 5;

    pointField& points = const_cast<pointField&>(mesh.points());

    plane approx=getPlane(points);
    linie theAxis(point(0,0,0),point(0,0,0));

    Info << "Plane of the grid: " << approx << "\n" << endl;

    const polyBoundaryMesh& patches = mesh.boundaryMesh();

    theAxis=linie(origin,origin+rotation);
    offset=0;
    
    Info << "The rotation-axis: " << theAxis << "\n" << endl;

    Info << "Creating wedge with an opening angle of " << wedgeAngle 
         << " degrees\n" << endl;

    changeCoordinates(
        mesh,
        approx,
        theAxis,
        offset,
        wedgeAngle,
	revolve
    );

    scalar distance=offset;

    const polyPatch &axisPatch=patches[patches.findPatchID(axisName_)];

    distance=getDistance(axisPatch,theAxis);

    Info << " Calculated Distance of " << axisName_ 
         << " to axis is " << distance << endl;

    if(distance>SMALL) { //Gill
        Info << "Distance of axis to patch is " << distance 
            << " -> not changing the patch type" << endl;
    }

    Info << "Splitting patch " << wedgeName << "\n" << endl;

    splitWedge(mesh,wedgeName,approx);

    Info << "Changing patch types " << endl;

    changeTypes(mesh, wedgeName, axisName_, distance>SMALL); 

    //Collapse axis patch to single line.

    scalar minLen(1e-8);
    scalar angle(4);

    scalar maxCos = Foam::cos(angle*constant::mathematical::pi/180.0);

    Info<< "Merging:" << nl
        << "    edges with length less than " << minLen << " meters" << nl
        << "    edges split by a point with edges in line to within " << angle
        << " degrees" << nl
        << endl;

    bool meshChanged = false;
  
    while (true)
    {
        const faceList& faces = mesh.faces();

        // Get all points on the boundary
        PackedBoolList boundaryPoint(mesh.nPoints());

        label nIntFaces = mesh.nInternalFaces();
        for (label faceI = nIntFaces; faceI < mesh.nFaces(); faceI++)
        {
            const face& f = faces[faceI];

            forAll(f, fp)
            {
                boundaryPoint.set(f[fp], 1);
            }
        }

        // Edge collapsing engine
        edgeCollapser collapser(mesh);

        // Collapse all edges that are too small.
        label nCollapsed =
            collapseSmallEdges
            (
                mesh,
                boundaryPoint,
                minLen,
                collapser
            );
        Info<< "Collapsing " << nCollapsed << " small edges" << endl;


        // Remove midpoints on straight edges.
        if (nCollapsed == 0)
        {
            nCollapsed = mergeEdges(mesh, maxCos, collapser);
            Info<< "Collapsing " << nCollapsed << " in line edges" << endl;
        }


        // Remove small sliver faces that can be collapsed to single edge
        if (nCollapsed == 0)
        {
            nCollapsed =
                collapseHighAspectFaces
                (
                    mesh,
                    boundaryPoint,
                    1E-9,       // factor of largest face area
                    5,          // factor between smallest and largest edge on
                                // face
                    collapser
                );
            Info<< "Collapsing " << nCollapsed
                << " small high aspect ratio faces" << endl;
        }

        // Simplify faces to quads wherever possible
        //if (nCollapsed == 0)
        //{
        //    nCollapsed =
        //        simplifyFaces
        //        (
        //            mesh,
        //            boundaryPoint,
        //            4,              // minimum size of face
        //            0.2,            // gap in edge lengths on face
        //            collapser
        //        );
        //    Info<< "Collapsing " << nCollapsed << " polygonal faces" << endl;
        //}


        if (nCollapsed == 0)
        {
            break;
        }

        directTopoChange meshMod(mesh);

        // Insert mesh refinement into directTopoChange.
        collapser.setRefinement(meshMod);

        // Do all changes
        Info<< "Morphing ..." << endl;

        autoPtr<mapPolyMesh> morphMap = meshMod.changeMesh(mesh, false);

        collapser.updateMesh(morphMap());

        if (morphMap().hasMotionPoints())
        {
            mesh.movePoints(morphMap().preMotionPoints());
        } 

        meshChanged = true;
    }

//    if (meshChanged) //Gill
//    {
        // Write resulting mesh
        mesh.setInstance(mesh.instance());

        Info<< "Writing collapsed mesh" << endl;

        mesh.write();
//    }
}


int main(int argc, char *argv[])
{
    argList::noParallel();
    argList::validArgs.append("Fluent mesh file");
    argList::validOptions.insert("scale", "scale factor");
    argList::validOptions.insert("writeCellSets", "");
    argList::validOptions.insert("writeFaceSets", "");
    argList::validOptions.insert("writeZones", "");

    argList args(argc, argv);

    if (!args.check())
    {
        FatalError.exit();
    }

    scalar scaleFactor = 1.0;
    args.optionReadIfPresent("scale", scaleFactor);

    bool writeCellSets = args.optionFound("writeCellSets");
    bool writeFaceSets = args.optionFound("writeFaceSets");
    bool writeZones = args.optionFound("writeZones");
    bool axiSymmetric = false;
    bool multiRegion = false;

#   include "createTime.H"

    fileName fluentFile(args.additionalArgs()[0]);
    fileName axisName("axis");
    std::ifstream fluentStream(fluentFile.c_str());

    if (!fluentStream)
    {
        FatalErrorIn("fluentToFoam::main(int argc, char *argv[])")
            << args.executable()
            << ": file " << fluentFile << " not found"
            << exit(FatalError);
    }

    //HR 22.11.09: These global variables do not get initialised.
    // Debug-version fails!
    patchTypeIDs.setSize(maxZoneID);
    patchNameIDs.setSize(maxZoneID);

    yyFlexLexer lexer(&fluentStream);
    while(lexer.yylex() != 0)
    {}

    Info<< "\n\nFINISHED LEXING\n\n\n";
    // Lookup table giving number of vertices given a fluent cell type ID
    // Currently not used.
    // label fluentModelNVertices[7] = {-1, 3, 4, 4, 8, 5, 6};

    // Lookup table giving number of vertices given a fluent cell type ID
    label fluentModelNFaces[7] = {-1, 3, 4, 4, 6, 5, 5};

    // Make a list of cell faces to be filled in for owner and neighbour

    labelListList cellFaces(nCells);

    labelList nFacesInCell(nCells, 0);

    forAll (cellFaces, celli)
    {
        cellFaces[celli].setSize(fluentModelNFaces[fluentCellModelID[celli] ]);
    }

    // fill in owner and neighbour

    forAll (owner, faceI)
    {
        if (owner[faceI] > -1)
        {
            label curCell = owner[faceI];

            if (nFacesInCell[curCell] >= cellFaces[curCell].size())
            {
                FatalErrorIn(args.executable())
                    << "Trying to add " << nFacesInCell[curCell] + 1
                    << "th face to a cell with " << cellFaces[curCell].size()
                    << " faces.  Face index: " << faceI
                    << " Current faces: " << cellFaces[curCell]
                    << abort(FatalError);
            }

            cellFaces[curCell][nFacesInCell[curCell] ] = faceI;

            nFacesInCell[curCell]++;
        }
    }

    forAll (neighbour, faceI)
    {
        if (neighbour[faceI] > -1)
        {
            label curCell = neighbour[faceI];

            if (nFacesInCell[curCell] >= cellFaces[curCell].size())
            {
                FatalErrorIn(args.executable())
                    << "Trying to add " << nFacesInCell[curCell] + 1
                    << "th face to a cell with " << cellFaces[curCell].size()
                    << " faces.  Face index: " << faceI
                    << " Current faces: " << cellFaces[curCell]
                    << abort(FatalError);
            }

            cellFaces[curCell][nFacesInCell[curCell] ] = faceI;

            nFacesInCell[curCell]++;
        }
    }

    // Construct shapes from face lists
    cellShapeList cellShapes(nCells);

    // Extrude 2-D mesh into 3-D

    Info<< "dimension of grid: " << dimensionOfGrid << endl;
    faceList frontAndBackFaces;

    if (dimensionOfGrid == 2)
    {
        const scalar extrusionFactor = 0.01;

        boundBox box(max(points), min(points));

        const scalar zOffset = extrusionFactor*box.mag();

        // two-dimensional grid. Extrude in z-direction
        Info<< "Grid is 2-D. Extruding in z-direction by: "
            << 2*zOffset << endl;

        pointField oldPoints = points;

        const label pointOffset = oldPoints.size();

        points.setSize(2*pointOffset);

        label nNewPoints = 0;

        // Note: In order for the owner-neighbour rules to be right, the
        // points given by Fluent need to represent the FRONT plane of the
        // geometry. Therefore, the extrusion will be in -z direction

        forAll (oldPoints, pointI)
        {
            points[nNewPoints] = oldPoints[pointI];

            points[nNewPoints].z() = zOffset;

            nNewPoints++;
        }

        forAll (oldPoints, pointI)
        {
            points[nNewPoints] = oldPoints[pointI];

            points[nNewPoints].z() = -zOffset;

            nNewPoints++;
        }

        // 2-D shape recognition
        Info<< "Creating shapes for 2-D cells"<< endl;

        // Set the number of empty faces
        frontAndBackFaces.setSize(2*nCells);

        forAll (fluentCellModelID, celli)
        {
            switch (fluentCellModelID[celli])
            {
                case 1:
                {
                    cellShapes[celli] =
                        cellShape
                        (
                            extrudedTriangleCellShape
                            (
                                celli,
                                cellFaces[celli],
                                faces,
                                owner,
                                neighbour,
                                pointOffset,
                                frontAndBackFaces
                            )
                        );
                }
                break;

                case 3:
                {
                    cellShapes[celli] =
                        cellShape
                        (
                            extrudedQuadCellShape
                            (
                                celli,
                                cellFaces[celli],
                                faces,
                                owner,
                                neighbour,
                                pointOffset,
                                frontAndBackFaces
                            )
                        );
                }
                break;

                default:
                {
                    FatalErrorIn(args.executable())
                        << "unrecognised  2-D cell shape: "
                        << fluentCellModelID[celli]
                        << abort(FatalError);
                }
            }
        }

        // Create new faces
        forAll (faces, faceI)
        {

            if (faces[faceI].size() != 2)
            {
                FatalErrorIn(args.executable())
                    << "fluentMeshToFoam: a 2-D face defined with "
                    << faces[faceI].size() << " points." << endl;
            }

            labelList& newFace = faces[faceI];

            newFace.setSize(4);

            newFace[2] = newFace[1] + pointOffset;

            newFace[3] = newFace[0] + pointOffset;
        }

        // Create new cells from 2-D shapes
    }
    else
    {
        // 3-D shape recognition
        Info<< "Creating shapes for 3-D cells"<< endl;
        forAll (fluentCellModelID, celli)
        {
            if
            (
                fluentCellModelID[celli] == 2    // tet
             || fluentCellModelID[celli] == 4    // hex
             || fluentCellModelID[celli] == 5    // pyramid
             || fluentCellModelID[celli] == 6    // prism
            )
            {
                cellShapes[celli] =
                    cellShape
                    (
                        create3DCellShape
                        (
                            celli,
                            cellFaces[celli],
                            faces,
                            owner,
                            neighbour,
                            fluentCellModelID[celli]
                        )
                    );
            }
            else
            {
                FatalErrorIn(args.executable())
                    << "unrecognised 3-D cell shape: "
                    << fluentCellModelID[celli]
                    << abort(FatalError);
            }
        }
    }

    // boundary faces are oriented such that the owner is zero and the face
    // area vector points into the domain. Turn them round before making patches
    // for Foam compatibility

    forAll (faces, faceI)
    {
        if (owner[faceI] == -1)
        {
            // reverse face
            labelList oldFace = faces[faceI];

            forAllReverse(oldFace, i)
            {
                faces[faceI][oldFace.size() - i - 1] =
                    oldFace[i];
            }
        }
    }


    //make patchless mesh before analysing boundaries

    faceListList patches(0);
    wordList patchNames(0);
    wordList patchTypes(0);
    word defaultFacesName = "defaultFaces";
    word defaultFacesType = emptyPolyPatch::typeName;
    wordList patchPhysicalTypes(0);

    // Scale the points

    points *= scaleFactor;

    Info<< "Building patch-less mesh..." << flush;

    polyMesh pShapeMesh
    (
        IOobject
        (
            polyMesh::defaultRegion,
            runTime.constant(),
            runTime
        ),
        xferMove(points),
        cellShapes,
        patches,
        patchNames,
        patchTypes,
        defaultFacesName,
        defaultFacesType,
        patchPhysicalTypes
    );

    //dont write mesh yet, otherwise preservePatchTypes will be broken
    //and zones wont be written
    //checkmesh done after patch addition as well
    Info<< "done." << endl;


    Info<< endl << "Building boundary and internal patches." << endl;
    //adding patches after mesh construction allows topological checks
    //on whether a patch is internal or external, something fluent
    //doesnt seem to mind

    // Make boundary patches

    SLList<label>::iterator faceGroupZoneIDIter = faceGroupZoneID.begin();
    SLList<label>::iterator faceGroupStartIndexIter =
        faceGroupStartIndex.begin();
    SLList<label>::iterator faceGroupEndIndexIter = faceGroupEndIndex.begin();

    // Note. Not all groups of faces will be boundary patches.
    // Take care on construction

    //2D needs extra space for frontAndBack faces
    if (dimensionOfGrid == 2)
    {
        patches.setSize(faceGroupZoneID.size()+1);
        patchNames.setSize(faceGroupZoneID.size()+1);
        patchTypes.setSize(faceGroupZoneID.size()+1);
        patchPhysicalTypes.setSize(faceGroupZoneID.size()+1);
    }
    else
    {
        patches.setSize(faceGroupZoneID.size());
        patchNames.setSize(faceGroupZoneID.size());
        patchTypes.setSize(faceGroupZoneID.size());
        patchPhysicalTypes.setSize(faceGroupZoneID.size());
    }

    label nPatches = 0;

    // Colate information for all patches (internal and external)

    // Create a file listing patch type for each zone

    label maxZoneID = 0;

    for
    (
        SLList<label>::iterator fgzIDIter = faceGroupZoneID.begin();
        fgzIDIter != faceGroupZoneID.end();
        ++fgzIDIter
    )
    {
        maxZoneID = max(maxZoneID, fgzIDIter());
    }

    Info << "maxZoneID: " << maxZoneID << endl;
    wordIOList zoneToPatchName
    (
        IOobject
        (
            "zoneToPatchName",
            runTime.constant(),
            pShapeMesh.meshSubDir,
            pShapeMesh,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        wordList(maxZoneID + 1, "unknown")
    );


    for
    (
        ;
        faceGroupZoneIDIter != faceGroupZoneID.end()
     && faceGroupStartIndexIter != faceGroupStartIndex.end()
     && faceGroupEndIndexIter != faceGroupEndIndex.end();
        ++faceGroupZoneIDIter,
        ++faceGroupStartIndexIter,
        ++faceGroupEndIndexIter
    )
    {
        // Get face type and name
        const word& curPatchType = patchTypeIDs[faceGroupZoneIDIter()];

        const word& curPatchName = patchNameIDs[faceGroupZoneIDIter()];

        Info<< "Creating patch " << nPatches
            << " for zone: " << faceGroupZoneIDIter()
            << " start: " << faceGroupStartIndexIter()
            << " end: " << faceGroupEndIndexIter()
            << " type: " << curPatchType << " name: " << curPatchName << endl;

        // Record zone index
        zoneToPatchName[faceGroupZoneIDIter()] = curPatchName;

        // Make patch labels
        label faceLabel = faceGroupStartIndexIter() - 1;

        faceList patchFaces(faceGroupEndIndexIter() - faceLabel);

        forAll (patchFaces, faceI)
        {
            if
            (
                faces[faceLabel].size() == 3
             || faces[faceLabel].size() == 4
            )
            {
                patchFaces[faceI] = face(faces[faceLabel]);
                faceLabel++;
            }
            else
            {
                FatalErrorIn("fluentToFoam::main(int argc, char *argv[])")
                    << "unrecognised face shape with "
                    << patchFaces[faceI].size() << " vertices"
                    << abort(FatalError);
            }
        }

        // Inlets and outlets
        if
        (
            curPatchType == "pressure"
         || curPatchType == "pressure-inlet"
         || curPatchType == "inlet-vent"
         || curPatchType == "intake-fan"
         || curPatchType == "pressure-outlet"
         || curPatchType == "exhaust-fan"
         || curPatchType == "outlet-vent"
         || curPatchType == "pressure-far-field"
         || curPatchType == "velocity-inlet"
         || curPatchType == "mass-flow-inlet"
         || curPatchType == "outflow"
         || curPatchType == "periodic"
         || curPatchType == "shadow"
        )
        {
            patches[nPatches] = patchFaces;
            patchTypes[nPatches] = polyPatch::typeName;
            patchNames[nPatches] = curPatchName;

            nPatches++;
        }
        else if (curPatchType == "interface" ) //interface boundaries
        {
            patches[nPatches] = patchFaces;
            patchTypes[nPatches] = wallPolyPatch::typeName;
            patchNames[nPatches] = curPatchName;

            nPatches++;
        }
        else if (curPatchType == "wall" ) //wall boundaries
        {
            patches[nPatches] = patchFaces;
            patchTypes[nPatches] = wallPolyPatch::typeName;
            patchNames[nPatches] = curPatchName;

            nPatches++;
        }
        else if
        (
            curPatchType == "symmetry"
        ) //symmetry planes
        {
            patches[nPatches] = patchFaces;
            patchTypes[nPatches] = symmetryPolyPatch::typeName;
            patchNames[nPatches] = curPatchName;

            nPatches++;
        }
        else if
        (
            curPatchType == "axis"
        ) //axisymmetric
        {
            patches[nPatches] = patchFaces;
            patchTypes[nPatches] = wallPolyPatch::typeName;
            patchNames[nPatches] = curPatchName;
            axisName = curPatchName;
            axiSymmetric = true;

            nPatches++;
        }
        else if
        (
            curPatchType == "interior"
         || curPatchType == "internal"
         || curPatchType == "solid"
         || curPatchType == "fan"
         || curPatchType == "radiator"
         || curPatchType == "porous-jump"
        ) //interior boundaries - will not be added as patches
        {
            patches[nPatches] = patchFaces;
            patchTypes[nPatches] = "internal";
            patchNames[nPatches] = curPatchName;

            nPatches++;
        }
        else if
        (
            curPatchType == ""
        ) // Unnamed face regions default to interior patches
        {
            Info<< "Patch " << faceGroupZoneIDIter()
                << ": Faces are defined but "
                << "not created as a zone." << endl
                << "Null specification is only valid for internal faces."
                << endl;

            patches[nPatches] = patchFaces;
            patchTypes[nPatches] = "internal";
            patchNames[nPatches] = curPatchName;

            nPatches++;
        }
        else //unknown face regions are not handled
        {
            FatalErrorIn(args.executable())
                << "fluent patch type " << curPatchType << " not recognised."
                << abort(FatalError);
        }
    }

    //add front and back boundaries for 2D meshes
    if (dimensionOfGrid == 2)
    {
        Info<< "Creating patch for front and back planes" << endl << endl;

        patches[nPatches] = frontAndBackFaces;
        patchTypes[nPatches] = emptyPolyPatch::typeName;
        patchNames[nPatches] = "frontAndBackPlanes";

        nPatches++;
    }

    //Now have all patch information,
    //check whether each patch is internal or external
    //and add boundaries to mesh
    //also write face sets of all patches
    patches.setSize(nPatches);
    patchTypes.setSize(nPatches);
    patchNames.setSize(nPatches);


    //old polyBoundary
    const polyBoundaryMesh& oPatches = pShapeMesh.boundaryMesh();
    // new patches.
    DynamicList<polyPatch*> newPatches(nPatches);

    // For every boundary face the old patch.
    labelList facePatchID(pShapeMesh.nFaces()-pShapeMesh.nInternalFaces(), -1);
    label cMeshFace = pShapeMesh.nInternalFaces();
    label nBoundaries = 0;

    forAll(patches, patchI)
    {
        const faceList& bFaces = patches[patchI];

        label sz = bFaces.size();
        labelList meshFaces(sz,-1);


        //make face set and write (seperate from rest for clarity)
        //internal and external Fluent boundaries
        {
            faceSet pFaceSet(pShapeMesh, patchNames[patchI], sz);

            forAll (bFaces, j)
            {
                const face& f = bFaces[j];
                label cMeshFace = findFace(pShapeMesh, f);
                meshFaces[j] = cMeshFace;
                pFaceSet.insert(cMeshFace);
            }
            if (writeFaceSets)
            {
                Info<< "Writing patch " << patchNames[patchI]
                    << " of size " << sz << " to faceSet." << endl;
                pFaceSet.write();
            }
        }


        //check if patch is internal
        //also check internal/external-ness of first patch face
        //internal faces cannot become foam boundaries
        //if a face is defined as internal but is actually external
        //it will be put in a default wall boundary
        //internal boundaries are simply ignored

        
        if
        (
            patchTypes[patchI] != "internal"
         && !pShapeMesh.isInternalFace(meshFaces[0])
        )
        {
            // First face is external and has valid non-internal type

            //check all faces for externalness just to be sure
            //and mark patch number to global list
            forAll(meshFaces, i)
            {
                label faceI = meshFaces[i];

                if (pShapeMesh.isInternalFace(faceI))
                {
                    FatalErrorIn(args.executable())
                        << "Face " << faceI << " on new patch "
                        << patchNames[patchI]
                        << " is not an external face of the mesh." << endl
                        << exit(FatalError);
                }

                if(facePatchID[faceI - pShapeMesh.nInternalFaces()]!= -1)
                {
                    FatalErrorIn(args.executable())
                        << "Face " << faceI << " on new patch "
                        << patchNames[patchI]
                        << " has already been marked for repatching to"
                        << " patch "
                        << facePatchID[faceI - pShapeMesh.nInternalFaces()]
                        << exit(FatalError);
                }
                facePatchID[faceI - pShapeMesh.nInternalFaces()] = nBoundaries;
            }

            // Add to boundary patch

            Info<< "Adding new patch " << patchNames[patchI]
                << " of type " << patchTypes[patchI]
                << " as patch " << nBoundaries << endl;

            // Add patch to new patch list
            newPatches.append
            (
                polyPatch::New
                (
                    patchTypes[patchI],
                    patchNames[patchI],
                    sz,
                    cMeshFace,
                    nBoundaries,
                    oPatches
                ).ptr()
            );

            nBoundaries++;
            cMeshFace += sz;
        }
        else
        {
            Info<< "Patch " << patchNames[patchI]
                 << " is internal to the mesh "
                 << " and is not being added to the boundary."
                 << endl;

        }
    }

    // Check for any remaining boundary faces
    // and add them to a default wall patch
    // this routine should generally not be invoked
    {
        DynamicList<label> defaultBoundaryFaces(facePatchID.size());
        forAll(facePatchID, idI)
        {
            if(facePatchID[idI] == -1)
            {
                defaultBoundaryFaces.append(idI);
                facePatchID[idI] = nBoundaries;
            }
        }
        defaultBoundaryFaces.shrink();

        if (defaultBoundaryFaces.size() != 0)
        {
            Warning << " fluent mesh has " << defaultBoundaryFaces.size()
                    << " undefined boundary faces." << endl
                    << " Adding undefined faces to new patch `default_wall`"
                    << endl;

            // Add patch to new patch list

            newPatches.append
            (
                polyPatch::New
                (
                    wallPolyPatch::typeName,
                    "default_wall",
                    defaultBoundaryFaces.size(),
                    cMeshFace,
                    nBoundaries,
                    oPatches
                ).ptr()
            );
            nBoundaries++;
            cMeshFace += defaultBoundaryFaces.size();
        }
    }

    newPatches.shrink();

    // Use facePatchIDs map to reorder boundary faces into compact regions

    repatchPolyTopoChanger repatcher(pShapeMesh);

    // Add new list of patches
    repatcher.changePatches(newPatches);

    // Change patch ids
    forAll(facePatchID, idI)
    {
        label faceI = idI + pShapeMesh.nInternalFaces();

        repatcher.changePatchID(faceI, facePatchID[idI]);
    }
    repatcher.repatch();

    // Set the precision of the points data to 10
    IOstream::defaultPrecision(10);


    // Zones
    // will write out cell zones and internal faces for those zones
    // note: zone boundary faces are not added to face zones
    // the names of boundaries bordering on cell zones are written to
    // a list containing the boundary name and cellzone it borders on
    // interior boundaries are handled via faceSets
    // cell zones will only be written if there is more than one

    if (writeZones && cellGroupZoneID.size()>1)
    {
        Info<< "Adding Zones" << endl;
        List<pointZone*> pz(0);

        label nrCellZones = cellGroupZoneID.size();
        List<cellZone*> cz(nrCellZones);

        // Make face zones for cell zones
        List<faceZone*> fz(nrCellZones);

        // List of patch names and the cellZone(s) they border
        // this is just an info file to make MRF easier to setup
        List<DynamicList<word> > boundaryZones
        (
            pShapeMesh.boundaryMesh().size()
        );

        const polyBoundaryMesh& bPatches = pShapeMesh.boundaryMesh();
        forAll(bPatches, pI)
        {
            boundaryZones[pI].append(bPatches[pI].name());
        }

        label cnt = 0;
        SLList<label>::iterator cg = cellGroupZoneID.begin();
        SLList<label>::iterator start = cellGroupStartIndex.begin();
        SLList<label>::iterator end = cellGroupEndIndex.begin();

        for (; cg != cellGroupZoneID.end(); ++cg, ++start, ++end)
        {
            const word& name = patchNameIDs[cg()];
            const word& type = patchTypeIDs[cg()];

            Info<< "Writing cell zone: " << name
                << " of type " << type << " starting at " << start() - 1
                << " ending at " << end() - 1 << " to cellSet." << endl;

            labelList cls(end() - start() + 1);

            // Mark zone cells, used for finding faces
            boolList zoneCell(pShapeMesh.nCells(), false);

            // Shift cell indices by 1
            label nr = 0;
            for (label celli = (start() - 1); celli < end(); celli++)
            {
                cls[nr] = celli;
                zoneCell[celli] = true;
                nr++;
            }

            cz[cnt] = new cellZone
            (
                name,
                cls,
                cnt,
                pShapeMesh.cellZones()
            );

            DynamicList<label> zoneFaces(pShapeMesh.nFaces());
            forAll(pShapeMesh.faceNeighbour(), faceI)
            {
                label nei = pShapeMesh.faceNeighbour()[faceI];
                label own = pShapeMesh.faceOwner()[faceI];
                if(nei != -1)
                {
                    if(zoneCell[nei] && zoneCell[own])
                    {
                        zoneFaces.append(faceI);
                    }
                }
            }
            zoneFaces.shrink();

            fz[cnt] = new faceZone
            (
                name,
                zoneFaces,
                boolList(zoneFaces.size(), false),
                cnt,
                pShapeMesh.faceZones()
            );

            // Add cell zones to patch zone list
            forAll(bPatches, pI)
            {
                const labelList& faceCells = bPatches[pI].faceCells();
                forAll(faceCells, fcI)
                {
                    if(zoneCell[faceCells[fcI] ])
                    {
                        boundaryZones[pI].append(name);
                        break;
                    }
                }
            }

            cnt++;
        }

        pShapeMesh.addZones(pz, fz, cz);

        forAll(bPatches, pI)
        {
            boundaryZones[pI].shrink();
        }

        fileName bczf
        (
            runTime.path()/runTime.constant()
           /"polyMesh"/"boundaryAdjacentCellZones"
        );

        OFstream boundaryCellZonesFile(bczf);
        forAll (boundaryZones, bzI)
        {
            forAll(boundaryZones[bzI], bzII)
            {
                boundaryCellZonesFile << boundaryZones[bzI][bzII] << " ";
            }

            boundaryCellZonesFile << endl;
        }
    }

    Info<< endl << "Writing mesh..." << flush;

    Info<< " to " << pShapeMesh.instance()/pShapeMesh.meshDir()
        << "  " << flush;

    pShapeMesh.setInstance(pShapeMesh.instance());
    pShapeMesh.write();

    // Write zone to patch name
    zoneToPatchName.write();
    Info<< "done." << endl << endl;

    // Write cellSets for Fluent regions
    // allows easy post-processing
    // set and zone functionality will be integrated some time
    // soon negating the need for double output
    if (writeCellSets)
    {
        if (cellGroupZoneID.size() > 0 )
        {
            Info<< "Writing cell sets" << endl;

            SLList<label>::iterator cg = cellGroupZoneID.begin();
            SLList<label>::iterator start = cellGroupStartIndex.begin();
            SLList<label>::iterator end = cellGroupEndIndex.begin();

            // Note: cellGroupXXX are all Fluent indices (starting at 1)
            // so offset before using.

            for (; cg != cellGroupZoneID.end(); ++cg, ++start, ++end)
            {
                const word& name=patchNameIDs[cg()];
                const word& type=patchTypeIDs[cg()];

                Info<< "Writing cell set: " << name
                    << " of type " << type << " starting at " << start() - 1
                    << " ending at " << end() - 1 << " to cellSet." << endl;

                cellSet internal(pShapeMesh, name, end() - start());

                // shift cell indizes by 1
                for(label celli=start() - 1; celli<=end() - 1; celli++)
                {
                    internal.insert(celli);
                }

                internal.write();
            }
        }
        else
        {
            Info<< "Only one cell group: no set written\n";
        }
    }

    if(axiSymmetric)
    {
        makeAxialMesh(pShapeMesh, axisName);
    }

    bool foundFluid = false;
    bool foundSolid = false;
    forAll(patchTypeIDs, zoneI)
    {
        if (patchTypeIDs[zoneI] == "fluid")
        {
            foundFluid = true;
        }
        else if (patchTypeIDs[zoneI] == "solid")
        {
            foundSolid = true;
        }

        if (foundFluid && foundSolid)
        {
            multiRegion = true;
            break;
        }
    }

    if (multiRegion)
    {
        // write 'setBatch' file for multi-region mesh //Gill
        fileName setBatch("setBatch");
        OFstream os(setBatch);
        foundFluid = false;

        forAll(patchTypeIDs, zoneI)
        {
            if (patchTypeIDs[zoneI] == "fluid" && !foundFluid)
            {
                os << "cellSet fluid new zoneToCell " << patchNameIDs[zoneI] 
                   << endl;
                foundFluid = true;
            }
            else if (patchTypeIDs[zoneI] == "fluid" && foundFluid)
            {
                os << "cellSet fluid add zoneToCell " << patchNameIDs[zoneI]
                   << endl;
            }
        }
        os << endl;
        forAll(patchTypeIDs, zoneI)
        {
            if (patchTypeIDs[zoneI] == "solid")
            {
                os << "cellSet " << patchNameIDs[zoneI] << " new zoneToCell " 
                   << patchNameIDs[zoneI]
                   << endl;
            }
        }

        // write 'regionProperties' dictionary for multi-region mesh //Gill
        fileName regionProperties(runTime.constant()/"regionProperties");
        OFstream osrp(regionProperties);

        IOobject::writeBanner(osrp)
            << "FoamFile\n{\n"
            << "    version     " << osrp.version() << ";\n"
            << "    format      " << osrp.format() << ";\n"
            << "    class       " << "dictionary" << ";\n"
            << "    location    " << "\"constant\"" << ";\n"
            << "    object      " << "regionProperties" << ";\n"
            << "}" << nl;
        IOobject::writeDivider(osrp) 
            << nl
            << "regions" 
            << nl << "(" 
            << nl << "    fluid    ( fluid )"
            << nl << "    solid    ( ";

        forAll(patchTypeIDs, zoneI)
        {
            if (patchTypeIDs[zoneI] == "solid")
            {
                osrp << patchNameIDs[zoneI] << " ";
            }
        }

        osrp << ")" << nl << ");" << nl;
        IOobject::writeEndDivider(osrp);
    }

    Info<< nl << "End" << endl;
    return 0;
}


/* --------------------------------------------------------------------------*\
    ------ End of fluentMeshToFoam.L
\* --------------------------------------------------------------------------*/
