/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Namespace
    Foam::diffusivityModels

Description
    A namespace for various incompressible diffusivityModel implementations.

Class
    Foam::diffusivityModel

Description
    An abstract base class for incompressible diffusivityModels.

    The strain rate is defined by:

        mag(symm(grad(U)))


SourceFiles
    diffusivityModel.C
    diffusivityModelNew.C

\*---------------------------------------------------------------------------*/

#ifndef diffusivityModel_H
#define diffusivityModel_H

#include "regIOobject.H"
#include "dictionary.H"
#include "volFieldsFwd.H"
#include "surfaceFieldsFwd.H"
#include "dimensionedScalar.H"
#include "runTimeSelectionTables.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
                Class diffusivityModel Declaration
\*---------------------------------------------------------------------------*/

class diffusivityModel
:
    public regIOobject
{

protected:

    // Protected data

        word name_;
        dictionary diffusivityProperties_;

        const volScalarField& scalar_;


    // Private Member Functions

        //- Disallow copy construct
        diffusivityModel(const diffusivityModel&);

        //- Disallow default bitwise assignment
        void operator=(const diffusivityModel&);


public:

    //- Runtime type information
    TypeName("diffusivityModel");


    // Declare run-time constructor selection table

        declareRunTimeSelectionTable
        (
            autoPtr,
            diffusivityModel,
            dictionary,
            (
                const word& name,
                const dictionary& diffusivityProperties,
                const volScalarField& scalar
            ),
            (name, diffusivityProperties, scalar)
        );


    // Selectors

        //- Return a reference to the selected diffusivity model
        static autoPtr<diffusivityModel> New
        (
            const word& name,
            const dictionary& diffusivityProperties,
            const volScalarField& scalar
        );


    // Constructors

        //- Construct from components
        diffusivityModel
        (
            const word& name,
            const dictionary& diffusivityProperties,
            const volScalarField& scalar
        );


    //- Destructor
    virtual ~diffusivityModel()
    {}


    // Member Functions

        //- Return the phase transport properties dictionary
        const dictionary& diffusivityProperties() const
        {
            return diffusivityProperties_;
        }

        //- Return the diffusivity
        virtual tmp<volScalarField> d() const = 0;

        //- Return the diffusivity for patch
        virtual tmp<scalarField> d(const label patchi) const = 0;

        //- Correct the diffusivity
        virtual void correct() = 0;

        //- Read transportProperties dictionary
        virtual bool read(const dictionary& diffusivityProperties) = 0;

        //- Default dummy write function
        virtual bool writeData(Ostream&) const
        {
            return true;
        }

};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
