/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | foam-extend: Open Source CFD
   \\    /   O peration     |
    \\  /    A nd           | For copyright notice see file Copyright
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of foam-extend.

    foam-extend is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation, either version 3 of the License, or (at your
    option) any later version.

    foam-extend is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with foam-extend.  If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::fv::VKLimitedGrad

Description
    VKLimitedGrad gradient scheme applied to a runTime selected base
    gradient scheme.

    The scalar limiter based on limiting the extrapolated face values
    between the maximum and minimum cell and cell neighbour values and is
    applied to the gradient in each face direction separately.

SourceFiles
    VKLimitedGrad.C

\*---------------------------------------------------------------------------*/

#ifndef VKLimitedGrad_H
#define VKLimitedGrad_H

#include "gradScheme.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace fv
{

/*---------------------------------------------------------------------------*\
                       Class VKLimitedGrad Declaration
\*---------------------------------------------------------------------------*/

template<class Type>
class VKLimitedGrad
:
    public fv::gradScheme<Type>
{
    // Private Data

        tmp<fv::gradScheme<Type> > basicGradScheme_;

        //- Limiter coefficient
        const scalar k_;

        //- Limiter epsilon coefficient
        const scalar K_;

    // Private Member Functions

        //- Disallow default bitwise copy construct
        VKLimitedGrad(const VKLimitedGrad&);

        //- Disallow default bitwise assignment
        void operator=(const VKLimitedGrad&);


public:

    //- RunTime type information
    TypeName("VKLimited");


    // Constructors

        //- Construct from mesh and schemeData
        VKLimitedGrad(const fvMesh& mesh, Istream& schemeData)
        :
            gradScheme<Type>(mesh),
            basicGradScheme_(fv::gradScheme<Type>::New(mesh, schemeData)),
            k_(readScalar(schemeData)),
            K_(readScalar(schemeData))
        {
            if (k_ < 0 || k_ > 1)
            {
                FatalIOErrorIn
                (
                    "VKLimitedGrad(const fvMesh& mesh, Istream& schemeData)",
                    schemeData
                )   << "Limiter coefficient = " << k_
                    << " should be >= 0 and <= 1"
                    << exit(FatalIOError);
            }
            if (K_ < 0)
            {
                FatalIOErrorIn
                (
                    "VKLimitedGrad(const fvMesh& mesh, Istream& schemeData)",
                    schemeData
                )   << "Limiter epsilon coefficient = " << K_
                    << " should be >= 0"
                    << exit(FatalIOError);
            }
        }


    // Member Functions

        static inline void limitFace
        (
            Type& limiter,
            const scalar& K,
            const scalar& deltah,
            const Type& maxDelta,
            const Type& minDelta,
            const Type& extrapolate
        );

        virtual tmp
        <
            GeometricField
            <typename outerProduct<vector, Type>::type, fvPatchField, volMesh>
        > calcGrad
        (
            const GeometricField<Type, fvPatchField, volMesh>& vsf,
            const word& name
        ) const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

template<>
inline void VKLimitedGrad<scalar>::limitFace
(
    scalar& limiter,
    const scalar& K,
    const scalar& deltah,
    const scalar& maxDelta,
    const scalar& minDelta,
    const scalar& extrapolate
)
{
    scalar sqrEpsilon = pow3(K)*pow3(deltah);

    if (extrapolate > 0)
    {
        scalar refDelta = extrapolate + 1e-12;

        limiter = 
            min
            (
                limiter,
                (sqr(maxDelta) + sqrEpsilon + 2.0*refDelta*maxDelta)/
                (sqr(maxDelta) + 2.0*sqr(refDelta)
                    + maxDelta*refDelta+sqrEpsilon)
            );
    }
    else if (extrapolate < 0)
    {
        scalar refDelta = extrapolate - 1e-12;

        limiter = 
            min
            (
                limiter,
                (sqr(minDelta) + sqrEpsilon + 2.0*refDelta*minDelta)/
                (sqr(minDelta) + 2.0*sqr(refDelta)
                    + minDelta*refDelta+sqrEpsilon)
            );
    }
    else
    {
        limiter = min(limiter, 1.0);
    }
}


template<class Type>
inline void VKLimitedGrad<Type>::limitFace
(
    Type& limiter,
    const scalar& K,
    const scalar& deltah,
    const Type& maxDelta,
    const Type& minDelta,
    const Type& extrapolate
)
{
    for(direction cmpt=0; cmpt<Type::nComponents; cmpt++)
    {
        VKLimitedGrad<scalar>::limitFace
        (
            limiter.component(cmpt),
            K,
            deltah,
            maxDelta.component(cmpt),
            minDelta.component(cmpt),
            extrapolate.component(cmpt)
        );
    }
}


// * * * * * * * * Template Member Function Specialisations  * * * * * * * * //

template<>
tmp<volVectorField> VKLimitedGrad<scalar>::calcGrad
(
    const volScalarField& vsf,
    const word& name
) const;


template<>
tmp<volTensorField> VKLimitedGrad<vector>::calcGrad
(
    const volVectorField& vsf,
    const word& name
) const;


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace fv

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
