/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::momentumWeighted

Description
    Interpolation scheme class which uses momentum equation's diagonal
	coefficients as weighting factors.

SourceFiles
    momentumWeighted.C

\*---------------------------------------------------------------------------*/

#ifndef momentumWeighted_H
#define momentumWeighted_H

#include "surfaceInterpolationScheme.H"
#include "volFields.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
                           Class momentumWeighted Declaration
\*---------------------------------------------------------------------------*/

template<class Type>
class momentumWeighted
:
    public surfaceInterpolationScheme<Type>
{
    // Private Data

        surfaceScalarField weights_;

    // Private Member Functions

        //- Disallow default bitwise assignment
        void operator=(const momentumWeighted&);


public:

    //- Runtime type information
    TypeName("momentumWeighted");


    // Constructors

        //- Construct from mesh
        momentumWeighted(const fvMesh& mesh)
        :
            surfaceInterpolationScheme<Type>(mesh),
            weights_(calcWeights())
        {}

        //- Construct from Istream
        momentumWeighted(const fvMesh& mesh, Istream&)
        :
            surfaceInterpolationScheme<Type>(mesh),
            weights_(calcWeights())
        {}

        //- Construct from faceFlux and Istream
        momentumWeighted
        (
            const fvMesh& mesh,
            const surfaceScalarField&,
            Istream&
        )
        :
            surfaceInterpolationScheme<Type>(mesh),
            weights_(calcWeights())
        {}


    // Member Functions

        //- Calculate the interpolation weighting factors
        tmp<surfaceScalarField> calcWeights() const
        {
            const fvMesh& mesh = this->mesh();

            tmp<volScalarField> AU 
            (
                mesh.lookupObject<volScalarField>("rAU")
            );

            tmp<surfaceScalarField> tw
            (
                new surfaceScalarField
                (
                    IOobject
                    (
                        "momentumWeightedWeights",
                        this->mesh().time().timeName(),
                        this->mesh()
                    ),
                    this->mesh(),
                    dimensionedScalar("1.0", dimless, 1.0)
                )
            );

            scalarField& w = tw().internalField();

    		const labelUList& own = mesh.owner();
    		const labelUList& nei = mesh.neighbour();

    		forAll(own, facei)
    		{
        		scalar AOwn = AU()[own[facei]]/mesh.V()[own[facei]];
        		scalar ANei = AU()[nei[facei]]/mesh.V()[nei[facei]];
        		w[facei] = AOwn/(AOwn + ANei);
    		}

    		surfaceScalarField::GeometricBoundaryField& bw 
				= tw().boundaryField();

    		forAll(bw, patchi)
    		{
        		const fvPatchScalarField& pAU = AU().boundaryField()[patchi];

        		if (pAU.coupled())
        		{
					const scalarField AUP  = pAU.patchInternalField();
					const scalarField AUN  = pAU.patchNeighbourField();

					forAll(pAU, facei)
					{
						label faceCelli = pAU.patch().faceCells()[facei];

						scalar AOwn = AUP[facei]/mesh.V()[faceCelli];
						scalar ANei = AUN[facei]/mesh.V()[faceCelli];
			
						bw[patchi][facei] = AOwn/(AOwn + ANei);
					}
        		}
    		}

            return tw;
        }

        //- Return the interpolation weighting factors
        tmp<surfaceScalarField> weights
        (
            const GeometricField<Type, fvPatchField, volMesh>&
        ) const
        {
            return weights_;
        }

        //- Return true if this scheme uses an explicit correction
        virtual bool corrected() const
        {
            return true;
        }

        //- Return the explicit correction to the face-interpolate
        virtual tmp<GeometricField<Type, fvsPatchField, surfaceMesh> >
        correction
        (
            const GeometricField<Type, fvPatchField, volMesh>&
        ) const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
