/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2014 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "pumpPowerCoeffs.H"
#include "dictionary.H"
#include "Time.H"
#include "Pstream.H"
#include "IOmanip.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineTypeNameAndDebug(pumpPowerCoeffs, 0);
}


// * * * * * * * * * * * * Protected Member Functions  * * * * * * * * * * * //

void Foam::pumpPowerCoeffs::writeFileHeader(const label i)
{
    if (i == 0)
    {
        // force coeff data

        writeHeader(file(i), "Power coefficients");
        writeHeaderValue(file(i), "shaftCenter", coordSys_.origin());
        writeHeaderValue(file(i), "shaftAxis", shaftAxis_);
        writeHeaderValue(file(i), "impellerSpeed", impellerSpeed_);
        writeHeaderValue(file(i), "volumeFlowRate", volumeFlowRate_);
        writeHeaderValue(file(i), "gravity", gravity_);
        writeCommented(file(i), "Time");
        writeTabbed(file(i), "Head[m]");
        writeTabbed(file(i), "ShaftPower[W]");
        writeTabbed(file(i), "Efficiency[%]");
    }
    else
    {
        FatalErrorIn("void Foam::pumpForces::writeFileHeader(const label)")
            << "Unhandled file index: " << i
            << abort(FatalError);
    }

    file(i)<< endl;
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::pumpPowerCoeffs::pumpPowerCoeffs
(
    const word& name,
    const objectRegistry& obr,
    const dictionary& dict,
    const bool loadFromFiles
)
:
    pumpForces(name, obr, dict, loadFromFiles, false),
    inletName_("none"),
    outletName_("none"),
    shaftAxis_(vector::zero),
    impellerSpeed_(0.0),
    volumeFlowRate_(0.0),
    gravity_(9.81)
{
    read(dict);

    Info<< endl;
}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::pumpPowerCoeffs::~pumpPowerCoeffs()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void Foam::pumpPowerCoeffs::read(const dictionary& dict)
{
    if (active_)
    {
        pumpForces::read(dict);

        dict.lookup("pumpInlet") >> inletName_;

        dict.lookup("pumpOutlet") >> outletName_;

        // Direction for torque
        dict.lookup("shaftAxis") >> shaftAxis_;

        // Impeller speed
        dict.lookup("impellerSpeed") >> impellerSpeed_;

        // Flow rate
        dict.lookup("volumeFlowRate") >> volumeFlowRate_;

        // Gravity
        dict.lookup("gravity") >> gravity_;
    }
}


void Foam::pumpPowerCoeffs::execute()
{
    // Do nothing - only valid on write
}


void Foam::pumpPowerCoeffs::end()
{
    // Do nothing - only valid on write
}


void Foam::pumpPowerCoeffs::timeSet()
{
    // Do nothing - only valid on write
}


Foam::scalar Foam::pumpPowerCoeffs::areaWeightedAveragePressure
(
    const word& patchName
)
{
    const volScalarField& p = obr_.lookupObject<volScalarField>(pName_);
    const fvMesh& mesh = p.mesh();

    const label patchI = mesh.boundaryMesh().findPatchID(patchName);

    scalar sumField = 0.0;

    scalar area = gSum(mesh.magSf().boundaryField()[patchI]);

    if (area > 0)
    {
        sumField = gSum
        (
            mesh.magSf().boundaryField()[patchI]
          * p.boundaryField()[patchI]
        ) / area;
    }

    return sumField;
}


void Foam::pumpPowerCoeffs::write()
{
    pumpForces::calcForcesMoment();

    if (!active_)
    {
        return;
    }

    scalar avgInletPressure = areaWeightedAveragePressure(inletName_);
    scalar avgOutletPressure = areaWeightedAveragePressure(outletName_);

    scalar head = (avgOutletPressure - avgInletPressure)/mag(gravity_);

    if (Pstream::master())
    {
        functionObjectFile::write();

        Field<vector> totMoment(moment_[0] + moment_[1] + moment_[2]);

        scalar ShaftPower = mag
        (
            impellerSpeed_*sum(totMoment & (shaftAxis_/mag(shaftAxis_)))
        );

        scalar Efficiency = mag
        (
            100*rhoRef_*volumeFlowRate_*head*mag(gravity_)/ShaftPower
        );

        file(0)
            << obr_.time().value() << tab << head << tab  << ShaftPower
            << tab << Efficiency << endl;

        Info(log_)<< type() << " " << " output:" << nl
            << "    Head       = " << head << " [m]" << nl
            << "    ShaftPower = " << ShaftPower << " [W]" << nl
            << "    Efficiency = " << Efficiency << " [%]" << endl;

        Info(log_)<< endl;
    }
}


// ************************************************************************* //
