/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2013 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "kEpsilon.H"
#include "addToRunTimeSelectionTable.H"

#include "backwardsCompatibilityWallFunctions.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace incompressible
{
namespace RASModels
{
namespace NEXT
{
// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

defineTypeNameAndDebug(kEpsilon, 0);
addToRunTimeSelectionTable(RASModel, kEpsilon, dictionary);

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

kEpsilon::kEpsilon
(
    const volVectorField& U,
    const surfaceScalarField& phi,
    transportModel& transport,
    const word& turbulenceModelName,
    const word& modelName
)
:
    RASModel(modelName, U, phi, transport, turbulenceModelName),

    Cmu_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cmu",
            coeffDict_,
            0.09
        )
    ),
    C1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "C1",
            coeffDict_,
            1.44
        )
    ),
    C2_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "C2",
            coeffDict_,
            1.92
        )
    ),
    sigmaEps_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "sigmaEps",
            coeffDict_,
            1.3
        )
    ),
    betaT_
    (
        "beta",
        dimless/dimTemperature,
        0
    ),
    Prt_
    (
        "Prt",
        dimless,
        0.85
    ),

    k_
    (
        IOobject
        (
            "k",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateK("k", mesh_)
    ),
    epsilon_
    (
        IOobject
        (
            "epsilon",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateEpsilon("epsilon", mesh_)
    ),
    nut_
    (
        IOobject
        (
            "nut",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateNut("nut", mesh_)
    ),
    Gb_
    (
        IOobject 
        (
            "Gb",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        dimensionedScalar
        (
            "zero",
            dimVelocity*dimAcceleration,
            pTraits<scalar>::zero
        )
    ),
    g_
    (
        IOobject 
        (
            "g",
            runTime_.constant(),
            mesh_,
            IOobject::READ_IF_PRESENT,
            IOobject::NO_WRITE
        ),
        dimensionedVector
        (
            "zero",
            dimAcceleration,
            pTraits<vector>::zero
        )
    )
{
    bound(k_, kMin_*10);
    bound(epsilon_, epsilonMin_*1e-5);

    nut_ = min(Cmu_*sqr(k_)/epsilon_, 1e5*nu());
    nut_.correctBoundaryConditions();

    printCoeffs();

    if (mesh_.foundObject<volScalarField>("T"))
    {
        const IOdictionary& transportProperties =
            mesh_.lookupObject<IOdictionary>("transportProperties");

        betaT_ = transportProperties.lookupOrDefault
        (
            "beta", 
            dimensionedScalar
            (
                "zero",
                dimless/dimTemperature,
                pTraits<scalar>::zero
            )
        );

        Prt_ = transportProperties.lookupOrDefault
        (
            "Prt",
            dimensionedScalar
            (
                "zero",
                dimless,
                0.85
            )
        );
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

tmp<volSymmTensorField> kEpsilon::R() const
{
    return tmp<volSymmTensorField>
    (
        new volSymmTensorField
        (
            IOobject
            (
                "R",
                runTime_.timeName(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            ((2.0/3.0)*I)*k_ - nut_*twoSymm(fvc::grad(U_)),
            k_.boundaryField().types()
        )
    );
}


tmp<volSymmTensorField> kEpsilon::devReff() const
{
    return tmp<volSymmTensorField>
    (
        new volSymmTensorField
        (
            IOobject
            (
                "devReff",
                runTime_.timeName(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
           -nuEff()*dev(twoSymm(fvc::grad(U_)))
        )
    );
}


tmp<fvVectorMatrix> kEpsilon::divDevReff(volVectorField& U) const
{
    return
    (
      - fvm::laplacian(nuEff(), U)
      //- fvc::div(nuEff()*dev(T(fvc::grad(U))))
      - fvc::div(nuEff()*T(fvc::grad(U))) //by Gill
    );
}


tmp<fvVectorMatrix> kEpsilon::divDevRhoReff
(
    const volScalarField& rho,
    volVectorField& U
) const
{
    volScalarField muEff("muEff", rho*nuEff());

    return
    (
      - fvm::laplacian(muEff, U)
      - fvc::div(muEff*dev(T(fvc::grad(U))))
    );
}


bool kEpsilon::read()
{
    if (RASModel::read())
    {
        Cmu_.readIfPresent(coeffDict());
        C1_.readIfPresent(coeffDict());
        C2_.readIfPresent(coeffDict());
        sigmaEps_.readIfPresent(coeffDict());

        return true;
    }
    else
    {
        return false;
    }
}


void kEpsilon::correct()
{
    RASModel::correct();

    if (!turbulence_)
    {
        return;
    }

    //volScalarField G(GName(), nuEff()*2*magSqr(symm(fvc::grad(U_))));
    volScalarField G(GName(), nut_*2*magSqr(symm(fvc::grad(U_))));

    // Production due to buoyancy
    if 
    (
        g_.value() != vector::zero 
     && mesh_.foundObject<volScalarField>("T")
    )
    {
        const volScalarField& T = mesh_.lookupObject<volScalarField>("T");
        Gb_ = (betaT_*nut_/Prt_)*(g_ & fvc::grad(T));
    }

    // Update epsilon and G at the wall
    epsilon_.boundaryField().updateCoeffs();

    volScalarField Nut = max(nut_, 1e-8*nu());

    // Dissipation equation
    tmp<fvScalarMatrix> epsEqn
    (
        fvm::ddt(epsilon_)
      + fvm::div(phi_, epsilon_)
      - fvm::laplacian(DepsilonEff(), epsilon_)
     ==
        0.5*C1_*G*sqrt(Cmu_)*sqrt(epsilon_)/sqrt(Nut)
      + 0.5*C2_*sqrt(Cmu_)*pow(epsilon_, 1.5)/sqrt(Nut) 
      - fvm::SuSp
        (
            1.5*C2_*sqrt(Cmu_)*sqrt(epsilon_)/sqrt(Nut) 
          - 0.5*C1_*G*sqrt(Cmu_)/sqrt(epsilon_)/sqrt(Nut),
            epsilon_
        )
    );

    epsEqn().relax();

    epsEqn().boundaryManipulate(epsilon_.boundaryField());

    solve(epsEqn);
    bound(epsilon_, epsilonMin_*1e-5);

    // Turbulent kinetic energy equation
    tmp<fvScalarMatrix> kEqn
    (
        fvm::ddt(k_)
      + fvm::div(phi_, k_)
      - fvm::laplacian(DkEff(), k_)
     ==
        G + Gb_ + Cmu_*sqr(k_)/Nut
      - fvm::SuSp(2*Cmu_*k_/Nut, k_)
    );

    kEqn().relax();
    
    solve(kEqn);
    bound(k_, kMin_*10);


    // Re-calculate viscosity
    nut_ = min(Cmu_*sqr(k_)/epsilon_, 1e5*nu());
    nut_.correctBoundaryConditions();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace NEXT
} // End namespace RASModels
} // End namespace incompressible
} // End namespace Foam

// ************************************************************************* //
